@inline function warpreduce(op, x, o, u = 0)
    while o > u
        x = op(x, shfl_down_sync(0xFFFFFFFF, x, o));
        o ÷= 2
    end
    x
end

@inline warpsum(x, o, u = 0) = warpreduce(+, x, o, u)
@inline warpprod(x, o, u = 0) = warpreduce(*, x, o, u)


function kernel_cpu!(R, X, P, I, ::Val{maxo}, ::Val{stride}) where {maxo, stride}
    for i in axes(I, 1), j in axes(P, 2)
        l = I[i,1]
        y = zero(eltype(R))

        for tidx in 1:stride
            x = fill(P[l+1,j], maxo ÷ stride + 1)
            for k = 2:l+1
                for o in 0:maxo÷stride
                    x[o + 1] *= X[I[i,k] + o * stride + tidx - 1, j]
                end
            end
            y += sum(x)
        end

        R[i, j] += y
    end

    return
end

function kernel!(results, fvalues, pfactors, imagelocations, images, polarization, background, ioff, roff, ::Val{maxo}, ::Val{stride}, ::Val{nsingle}) where {maxo, stride, nsingle}
    tidx = threadIdx().x; dimx = maxo ÷ stride + 1
    i = (blockIdx().y - one(blockIdx().y)) * blockDim().y + threadIdx().y + ioff
    j = (blockIdx().z - one(blockIdx().z)) * blockDim().z + threadIdx().z + roff

    fvalues = CUDA.Const(fvalues); imagelocations = CUDA.Const(imagelocations); images = CUDA.Const(images); pfactors = CUDA.Const(pfactors)
    polarization = isnothing(polarization) ? nothing : CUDA.Const(polarization); 
    background = isnothing(background) ? nothing : CUDA.Const(background)

    @inbounds if i <= length(imagelocations)
        imageindex = imagelocations[i]
        imagelength = images[imageindex]
        
        x = MVector{dimx, eltype(pfactors)}(undef)
        for o in 1:dimx
            x[o] = pfactors[imagelength + one(imagelength), j, ndims(pfactors) == 3 ? (o - 1) * stride + tidx : 1]
        end

        @inline function imageproduct!(x, k1, k2)
            @inbounds for k in k1:k2
                p = images[imageindex + k]
                for o in 1:dimx
                    intensity = fvalues[p + (o - one(o)) * stride + tidx - 1, j]
                    !isnothing(polarization) && (intensity *= polarization[p])
                    !isnothing(background) && (intensity += background[p])
                    x[o] *= intensity
                end
            end
        end

        if eltype(pfactors) == eltype(fvalues)
            imageproduct!(x, UInt32(1), imagelength)
        else
            xt = ones(MVector{dimx, eltype(fvalues)})
            k1 = UInt32(1)
            while k1 <= imagelength
                imageproduct!(xt, k1, min(k1 + nsingle - one(k1), imagelength))
                x .*= xt
                xt .= one(eltype(xt))
                k1 += nsingle
            end
        end

        cx = convert(MVector{length(x), eltype(results)}, x)
        y = warpsum(sum(cx), stride ÷ 2)

        if tidx == 1
            CUDA.@atomic results[i, j - roff] += y
        end
    end

    return
end

function applyKernel!(results, fvalues::CuArray, pfactors, imagelocations, images, maxoffset, rrange, polarization = nothing, background = nothing)
    dimx = min(8, maxoffset + 1)
    dimy = 512 ÷ dimx
    
    for ir in partitionrange(1 : ceil(Int, length(imagelocations) / dimy), maxl = 65535)
        blocks = (1, length(ir), length(rrange))
        ioff = UInt32((minimum(ir) - 1) * dimy)
        roff = UInt32(minimum(rrange) - 1)
        # println(CUDA.registers(@cuda name="main_kernel" threads=(dimx,dimy,1) blocks=blocks kernel!(results, fvalues, pfactors, imagelocations, images, polarization, background, ioff, roff, Val(maxoffset), Val(dimx), Val(32))))
        
        @cuda name="main_kernel" always_inline=true threads=(dimx,dimy,1) blocks=blocks kernel!(results, fvalues, pfactors, imagelocations, images, polarization, background, ioff, roff, Val(Int32(maxoffset)), Val(Int32(dimx)), Val(UInt32(50)))
    end
end

function applyKernel!(R, X, P, I, maxoffset)
    kernel_cpu!(R, X, P, I, Val(maxoffset), Val(8))
end
